function [p en]=runQCsimulation(varargin)
  %% Leader-follower main simulation file
  %
  % Simulation description: 
  % Basing on options structure the controller is created, using the
  % identified linear Quadcopter closed loop model, and a simulation in a
  % given scenario is performed, retuning the performance.
  %
  % Controller indces are:
  %
  % *  1 -> independent
  % *  2 -> decentralized output-feedback
  % *  3 -> centralized MPC
  % *  4 -> Decentralized MPC
  %
  defaults.controller=1; 
  defaults.videoname=[]; % if not empty (e.g. 'my_video_name') saves the output video with this name
  defaults.randx0=0;
  defaults.runIdentification=0;
  % Let the formation be defined by the following deviation w.r.t. the leader
  % position
  %       dx dy dz
  defaults.delta1=[5 -5 0]';
  defaults.delta2=[5 5 0]';  

  if nargin<1
    options=[];
  else
    options=varargin{1};
  end
  
  if isempty(options) || ~isstruct(options)
    clear options
    options=defaults;
  else
    fn=fieldnames(defaults);
    for i=1:length(fn)
      if ~isfield(options,fn{i})
        options=setfield(options,fn{i},getfield(defaults,fn{i}));
      end
    end
  end

  delta1=options.delta1;
  delta2=options.delta2;

  % Sampling time
  Ts=0.1;

  % Init controllers
  dummy=ss(zeros(3),eye(3),eye(3),zeros(3),Ts);
  big_dummy=append(dummy,dummy,dummy);
  dL=dlincon(big_dummy,'track');
  lK=zeros(3);
  K=zeros(9,9);



  %% Controllers definition

  % Identify the linear closed loop model: use the best trade-off between
  % complexity and performance
  if options.runIdentification
    try
      model;
    catch ee
      model=identifyQCmodel(4);
    end
  else
    load identifiedModel.mat
  end

  sys=model;
  sTs=10*Ts;
  sys=d2d(sys,sTs);

  % update DCgain to be unitary
  sys.b = sys.b * inv(dcgain(sys));

  msys = append(sys,sys,sys);

  nu=size(sys.b,2);

  Q=eye(size(sys.c,1));
  R=1e0*eye(nu);

  bQ=blkdiag(Q,2*Q,2*Q);
  bR=blkdiag(R,R,R);


  switch options.controller
    case 1
      lK=LMIout_cvx(sys.a,sys.b,sys.c,Q);
    case 2
      K=LMIout_cvx(msys.a,msys.b,msys.c,bQ);
    case {3,4}
      nu=size(msys.b,2);

      clear cost interval
      cost.S=bQ;
      cost.T=bR;
      cost.rho=inf;
      interval.N=10;
      interval.Nu=10;
      limits.dumin=-10*sTs*ones(nu,1);
      limits.dumax=-limits.dumin;
      limits.umin=-inf*ones(nu,1);
      limits.umax=-limits.umin;

      dec=[];
      nx=size(msys.a,1)/3;
      for i=1:3
        dec(i).x=[1+(i-1)*nx:i*nx];
        dec(i).y=[1+(i-1)*3:i*3];
        dec(i).u=[1+(i-1)*3:i*3];
        dec(i).applied=[1+(i-1)*3:i*3];
      end

      dL=dlincon(msys,'track',dec,cost,interval,limits);
      if options.controller==3
        dL=dL.setDefaultMode('global');
      else
        dL=dL.setDefaultMode('Dglobal');
      end
  end


  %% Simulation setup
  % simulation time
  Tsim=20;
  t=0:Ts:Tsim;
  Ns=length(t);

  % Define simulation initial conditions of leader and followers
  if options.randx0
    lx0=[1 1 0]'+10*rand(3,1);
    f1x0=[1 1 0]'+10*rand(3,1);
    f2x0=[1 1 0]'+10*rand(3,1);
    lx0= [zeros(3,1); lx0; zeros(6,1)];
    f1x0= [zeros(3,1); f1x0; zeros(6,1)];
    f2x0= [zeros(3,1); f2x0; zeros(6,1)];
  else
    lx0= [zeros(3,1);   10;10;0;  zeros(6,1)];
    f1x0= [zeros(3,1);  5;5;0;  zeros(6,1)];
    f2x0= [zeros(3,1);  5;0;10;  zeros(6,1)];
  end

    
  % Define Leader reference
  clear r
  nn=floor(Ns/2);
  r=[...
    [50*ones(1,nn) 10*ones(1,nn+1)];...
    30*ones(1,Ns);...
    [0*ones(1,nn) 20*ones(1,nn+1)]...
    ];
  % setup fromWorkspace Simulink block
  ref.time=t;
  ref.signals.values=r';
  ref.signals.dimension=3;
  
  % Load the MPC controller for the closed loop of each quadcopter (parameter
  % of the Simulink library block)
  evalin('base','load MPC_controller;');
  assignin('base','lK',lK);
  assignin('base','K',K);
  assignin('base','dL',dL);  
  assignin('base','ref',ref);    
  assignin('base','lx0',lx0);
  assignin('base','f1x0',f1x0);
  assignin('base','f2x0',f2x0); 
  assignin('base','delta1',delta1);
  assignin('base','delta2',delta2);   
  assignin('base','options',options);
  assignin('base','Ts',Ts);
  assignin('base','sTs',sTs);

  
  % perform simulation
  sim('LeaderFollower2',Tsim);
  disp('Simultion concluded');

  %% Plot results
  % plot in the same 3D axis all reference and trajectories
  pos{1}=r;
  pos{2}=lx.signals.values';
  pos{3}=f1x.signals.values';
  pos{4}=f2x.signals.values';

  plotTrajectories(pos,options.videoname);
  
  assignin('base','pos',pos);
  

  %% Compute performances
  n=size(pos{2},2);
  posf1=pos{1}+delta1*ones(1,n);
  posf2=pos{1}+delta2*ones(1,n);
  x=[pos{2}-pos{1};pos{3}-posf1;pos{4}-posf2];
  
  figure;
  plot(x');
  title('All quadcopter position deviations');
  
  p = performanceIndex(x,bQ);

  %% Compute energy consumption

  tot_l_en=sum(l_en(end,:));
  tot_f1_en=sum(f1_en(end,:));
  tot_f2_en=sum(f2_en(end,:));

  % absolute energy consumption
  en_abs=tot_l_en + tot_f1_en + tot_f2_en; 

  assignin('base','tot_l_en',tot_l_en);
  assignin('base','tot_f1_en',tot_f1_en);
  assignin('base','tot_f1_en',tot_f2_en);

  assignin('base','tot_en',tot_en);

  % lower bound energy value is 847.3276 obtained from 20s hovering
  lb_en=3*847.3276;
  % percent compared to lower bound
  en=en_abs*(100/lb_en);

end


